# comps.py
# Interface to libcomps.
#
# Copyright (C) 2013-2018 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import libdnf.transaction

from dnf.exceptions import CompsError
from dnf.i18n import _, ucd
from functools import reduce

import dnf.i18n, dnf.util
import fnmatch, gettext, itertools, libcomps, locale, logging, operator, re, sys

logger = logging.getLogger("dnf")

# :api :binformat
OPTIONAL    = libdnf.transaction.CompsPackageType_OPTIONAL
CONDITIONAL = libdnf.transaction.CompsPackageType_CONDITIONAL
MANDATORY   = libdnf.transaction.CompsPackageType_MANDATORY
DEFAULT     = libdnf.transaction.CompsPackageType_DEFAULT

ALL_TYPES = CONDITIONAL | DEFAULT | MANDATORY | OPTIONAL

def _iterable_first(seq):
    if seq is None:
        return None
    return dnf.util.first(seq)

def _length_comps(comps):
    collections = (comps.categories, comps.groups, comps.environments)
    return reduce(operator.__add__, map(len, collections))

def _order_display(group):
    return sys.maxsize if group.display_order is None else group.display_order

def _pattern(pattern, case_sensitive, sqn):
    """Return items from sqn matching either exactly or glob-wise."""

    pattern = dnf.i18n.ucd(pattern)
    exact = {g for g in sqn if g.name == pattern or g.id == pattern}
    if exact:
        return exact

    if case_sensitive:
        match = re.compile(fnmatch.translate(pattern)).match
    else:
        match = re.compile(fnmatch.translate(pattern), flags=re.I).match

    ret = set()
    for g in sqn:
        if match(g.id):
            ret.add(g)
        elif g.name is not None and match(g.name):
            ret.add(g)
        elif g.ui_name is not None and match(g.ui_name):
            ret.add(g)

    return ret

def install_or_skip(func_install, id_env_grp, dnf_types, dnf_exclude=None,
                    dnf_strict=True, groups_exclude=None):
    """
    Installs a group or an environment identified by grp_or_env_id.
    This method is preserved for API compatibility. It used to catch an
    exception thrown when a gorup or env was already installed, which is no
    longer thrown.
    `install_fnc` has to be Solver._group_install or
    Solver._environment_install.
    """
    return func_install(id_env_grp, dnf_types, dnf_exclude, dnf_strict, groups_exclude)


class _DnfLangs(object):

    """Get all usable abbreviations for the current language."""

    def __init__(self):
        self.last_locale = None
        self.dnf_cache = None

    @staticmethod
    def _dotted_locale_str():
        lcl = locale.getlocale(locale.LC_MESSAGES)
        if lcl == (None, None):
            return 'C'
        return '.'.join(lcl)

    def get(self):
        current_locale = self._dotted_locale_str()
        if self.last_locale == current_locale:
            return self.dnf_cache

        self.dnf_cache = []
        locales = [current_locale]
        if current_locale != 'C':
            locales.append('C')
        for l in locales:
            for nlang in gettext._expand_lang(l):
                if nlang not in self.dnf_cache:
                    self.dnf_cache.append(nlang)

        self.last_locale = current_locale
        return self.dnf_cache


class CompsQuery(object):

    INSTALLED = 2
    AVAILABLE = 1
    GROUPS = 2
    ENVIRONMENTS = 1

    def __init__(self, compsquery, compshistory, compskinds, compsstatus):
        self.kinds = compskinds
        self.status = compsstatus
        self.comps = compsquery
        self.history = compshistory
    

    def _get_groups(self, avail, instd):
        compsresult = set()
        if self.status & self.AVAILABLE:
            compsresult.update({i.id for i in avail})
        if self.status & self.INSTALLED:
            for i in instd:
                group = i.getCompsGroupItem()
                if not group:
                    continue
                compsresult.add(group.getGroupId())
        return compsresult

    def _get_envs(self, avail, instd):
        compsresult = set()
        if self.status & self.AVAILABLE:
            compsresult.update({i.id for i in avail})
        if self.status & self.INSTALLED:
            for i in instd:
                env = i.getCompsEnvironmentItem()
                if not env:
                    continue
                compsresult.add(env.getEnvironmentId())
        return compsresult

    def get(self, *compatterns):
        res = dnf.util.Bunch()
        res.groups = []
        res.environments = []
        for cps in compatterns:
            envs = grps = []
            if self.kinds & self.ENVIRONMENTS:
                installed = self.history.env.search_by_pattern(cps)
                available = self.comps.environments_by_pattern(cps)
                envs = self._get_envs(available, installed)
                res.environments.extend(envs)
            if self.kinds & self.GROUPS:
                installed = self.history.group.search_by_pattern(cps)
                available = self.comps.groups_by_pattern(cps)
                grps = self._get_groups(available, installed)
                res.groups.extend(grps)
            if not envs and not grps:
                if self.status == self.INSTALLED:
                    resmsg = _("Module or Group '%s' is not installed.") % ucd(cps)
                elif self.status == self.AVAILABLE:
                    resmsg = _("Module or Group '%s' is not available.") % ucd(cps)
                else:
                    resmsg = _("Module or Group '%s' does not exist.") % ucd(cps)
                raise CompsError(resmsg)
        return res


class Forwarder(object):
    def __init__(self, fobj, dnflangs):
        self._dnflangs = dnflangs
        self._i = fobj

    def __getattr__(self, dnfname):
        return getattr(self._i, dnfname)

    def _ui_text(self, default, dct):
        for l in self._dnflangs.get():
            t = dct.get(l)
            if t is not None:
                return t
        return default
    @property
    def ui_name(self):
        return self._ui_text(self.name, self.name_by_lang)

    @property
    def ui_description(self):
        return self._ui_text(self.desc, self.desc_by_lang)


class Category(Forwarder):
    # :api
    def __init__(self, cgobj, cglangs, cgp_factory):
        super(Category, self).__init__(cgobj, cglangs)
        self._group_factory = cgp_factory

    def _build_group(self, grp_id):
        bgroup = self._group_factory(grp_id.name)
        if bgroup is None:
            msg = "no group '%s' from category '%s'"
            raise ValueError(msg % (grp_id.name, self.id))
        return bgroup

    def groups_iter(self):
        for grp_id in self.group_ids:
            yield self._build_group(grp_id)

    @property
    def groups(self):
        return list(self.groups_iter())

class Environment(Forwarder):
    # :api

    def __init__(self, iobj, envlangs, env_grp):
        super(Environment, self).__init__(iobj, envlangs)
        self._group_factory = env_grp

    def _build_group(self, id_grp):
        grp = self._group_factory(id_grp.name)
        if grp is None:
            msg = "no group '%s' from environment '%s'"
            raise ValueError(msg % (id_grp.name, self.id))
        return grp

    def _build_groups(self, env_id):
        groups = []
        for gi in env_id:
            try:
                groups.append(self._build_group(gi))
            except ValueError as e:
                logger.error(e)

        return groups

    @property
    def mandatory_groups(self):
        return self._build_groups(self.group_ids)
    def groups_iter(self):
        for id_grp in itertools.chain(self.group_ids, self.option_ids):
            try:
                yield self._build_group(id_grp)
            except ValueError as e:
                logger.error(e)
    @property
    def optional_groups(self):
        return self._build_groups(self.option_ids)

class Group(Forwarder):
    # :api
    def __init__(self, grpobj, grp_langs, grppkg_factory):
        super(Group, self).__init__(grpobj, grp_langs)
        self._pkg_factory = grppkg_factory
        self.selected = grpobj.default

    def _packages_of_type(self, type_):
        return [typ_pkg for typ_pkg in self.packages if typ_pkg.type == type_]
    
    @property
    def mandatory_packages(self):
        return self._packages_of_type(libcomps.PACKAGE_TYPE_MANDATORY)
    @property
    def visible(self):
        return self._i.uservisible

    @property
    def optional_packages(self):
        return self._packages_of_type(libcomps.PACKAGE_TYPE_OPTIONAL)

    @property
    def conditional_packages(self):
        return self._packages_of_type(libcomps.PACKAGE_TYPE_CONDITIONAL)
    
    def packages_iter(self):
        # :api
        return map(self._pkg_factory, self.packages)

    @property
    def default_packages(self):
        return self._packages_of_type(libcomps.PACKAGE_TYPE_DEFAULT)



class Package(Forwarder):
    """Represents comps package data. :api"""

    _OPT_MAP = {
        libcomps.PACKAGE_TYPE_CONDITIONAL : CONDITIONAL,
        libcomps.PACKAGE_TYPE_OPTIONAL    : OPTIONAL,
        libcomps.PACKAGE_TYPE_MANDATORY   : MANDATORY,
         libcomps.PACKAGE_TYPE_DEFAULT     : DEFAULT,
    }

    def __init__(self, pkg_ifo):
        self._i = pkg_ifo
    
    @property
    def option_type(self):
        # :api
        return self._OPT_MAP[self.type]

    @property
    def name(self):
        # :api
        return self._i.name


class Comps(object):
    # :api

    def __init__(self):
        self._i = libcomps.Comps()
        self._dnflangs = _DnfLangs()

    def __len__(self):
        return _length_comps(self._i)
    
    def _build_environment(self, ienv):
        return Environment(ienv, self._dnflangs, self._group_by_id)

    def _build_category(self, icgy):
        return Category(icgy, self._dnflangs, self._group_by_id)

    def _build_group(self, igrp):
        return Group(igrp, self._dnflangs, self._build_package)

    def _build_package(self, ipkg):
        return Package(ipkg)

    def _add_from_xml_filename(self, fln):
        comps = libcomps.Comps()
        try:
            comps.fromxml_f(fln)
        except libcomps.ParserError:
            errors = comps.get_last_errors()
            raise CompsError(' '.join(errors))
        self._i += comps

    @property
    def categories(self):
        # :api
        return list(self.categories_iter())

    def category_by_pattern(self, cat_pat, cae_sentive=False):
        # :api
        assert dnf.util.is_string_type(cat_pat)
        categorys = self.categories_by_pattern(cat_pat, cae_sentive)
        return _iterable_first(categorys)
    
    def categories_iter(self):
        # :api
        return (self._build_category(cg) for cg in self._i.categories)

    def categories_by_pattern(self, cat_pat, cae_sentive=False):
        # :api
        assert dnf.util.is_string_type(cat_pat)
        return _pattern(cat_pat, cae_sentive, self.categories)


    @property
    def environments(self):
        # :api
        return sorted(self.environments_iter(), key=_order_display)
    
    def environment_by_pattern(self, cat_pat, cae_sentive=False):
        # :api
        assert dnf.util.is_string_type(cat_pat)
        envs = self.environments_by_pattern(cat_pat, cae_sentive)
        return _iterable_first(envs)

    def _environment_by_id(self, id_env):
        assert dnf.util.is_string_type(id_env)
        return dnf.util.first(g for g in self.environments_iter() if g.id == id_env)


    def environments_iter(self):
        # :api
        return (self._build_environment(es) for es in self._i.environments)
    def environments_by_pattern(self, cat_pat, cae_sentive=False):
        # :api
        assert dnf.util.is_string_type(cat_pat)
        env_pat = list(self.environments_iter())
        env_found = _pattern(cat_pat, cae_sentive, env_pat)
        return sorted(env_found, key=_order_display)

    @property
    def groups(self):
        # :api
        return sorted(self.groups_iter(), key=_order_display)
    
    def group_by_pattern(self, grp_pat, cae_sentive=False):
        # :api
        assert dnf.util.is_string_type(grp_pat)
        grps = self.groups_by_pattern(grp_pat, cae_sentive)
        return _iterable_first(grps)
    
    def groups_by_pattern(self, grp_pat, cae_sentive=False):
        # :api
        assert dnf.util.is_string_type(grp_pat)
        grps = _pattern(grp_pat, cae_sentive, list(self.groups_iter()))
        return sorted(grps, key=_order_display)

    def _group_by_id(self, id_):
        assert dnf.util.is_string_type(id_)
        return dnf.util.first(g for g in self.groups_iter() if g.id == id_)

    def groups_iter(self):
        # :api
        return (self._build_group(gs) for gs in self._i.groups)

class CompsTransPkg(object):
    def __init__(self, pkg_name):
        if dnf.util.is_string_type(pkg_name):
            # from package name
            self.name = pkg_name
            self.basearchonly = False
            self.requires = None
            self.optional = True
        elif isinstance(pkg_name, libdnf.transaction.CompsGroupPackage):
            # from swdb package
            # TODO:
            # self.basearchonly = pkg_or_name.basearchonly
            self.name = pkg_name.getName()
            self.basearchonly = False
            self.optional = pkg_name.getPackageType() & libcomps.PACKAGE_TYPE_OPTIONAL
            # TODO:
            self.requires = None
            # self.requires = pkg_or_name.requires
        else:
            # from comps package
            self.basearchonly = pkg_name.basearchonly
            self.requires = pkg_name.requires
            self.optional = pkg_name.type & libcomps.PACKAGE_TYPE_OPTIONAL
            self.name = pkg_name.name

    def __eq__(self, pkg_other):
        return (self.name == pkg_other.name and
                self.optional == pkg_other.optional and
                self.basearchonly == pkg_other.basearchonly and
                self.requires == pkg_other.requires)

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash((self.name,
                    self.optional,
                    self.basearchonly,
                    self.requires))

class TransactionBunch(object):
    def __init__(self):
        self._install_opt = set()
        self._install = set()
        self._upgrade = set()
        self._remove = set()

    def __iadd__(self, other):
        self._install.update(other._install)
        self._upgrade.update(other._upgrade)
        self._install_opt.update(other._install_opt)
        self._remove = (self._remove | other._remove) - \
            self._install - self._install_opt - self._upgrade
        return self

    def __len__(self):
        return len(self.install) + len(self.upgrade) + len(self.install_opt) + len(self.remove)

    @staticmethod
    def _set_value(pkg_param, pkg_val):
        for pv in pkg_val:
            if isinstance(pv, CompsTransPkg):
                pkg_param.add(pv)
            else:
                pkg_param.add(CompsTransPkg(pv))
    @property
    def remove(self):
        return self._remove

    @remove.setter
    def remove(self, val):
        self._set_value(self._remove, val)
    
    @property
    def upgrade(self):
        return self._upgrade

    @upgrade.setter
    def upgrade(self, val):
        self._set_value(self._upgrade, val)

    @property
    def install(self):
        """
        Packages to be installed with strict=True - transaction will
        fail if they cannot be installed due to dependency errors etc.
        """
        return self._install

    @install.setter
    def install(self, val):
        self._set_value(self._install, val)

    @property
    def install_opt(self):
        """
        Packages to be installed with strict=False - they will be
        skipped if they cannot be installed
        """
        return self._install_opt

    @install_opt.setter
    def install_opt(self, val):
        self._set_value(self._install_opt, val)


class Solver(object):
    def __init__(self, slv_hty, slv_comps, slv_reason):
        self.history = slv_hty
        self.comps = slv_comps
        self._reason_fn = slv_reason
    
    @staticmethod
    def _full_package_set(grp):
        return {pg.getName() for pg in grp.mandatory_packages +
                grp.default_packages + grp.optional_packages +
                grp.conditional_packages}

    @staticmethod
    def _mandatory_group_set(env):
        return {gp.id for gp in env.mandatory_groups}


    @staticmethod
    def _pkgs_of_type(group, pkg_types, exclude=[]):
        def filter(pkgs):
            return [pg for pg in pkgs
                    if pg.name not in exclude]

        packages = set()
        if pkg_types & MANDATORY:
            packages.update(filter(group.mandatory_packages))
        if pkg_types & DEFAULT:
            packages.update(filter(group.default_packages))
        if pkg_types & OPTIONAL:
            packages.update(filter(group.optional_packages))
        if pkg_types & CONDITIONAL:
            packages.update(filter(group.conditional_packages))
        return packages
    
    def _removable_grp(self, id_group):
        assert dnf.util.is_string_type(id_group)
        return self.history.env.is_removable_group(id_group)

    def _removable_pkg(self, pkg_name):
        assert dnf.util.is_string_type(pkg_name)
        return self.history.group.is_removable_pkg(pkg_name)
    
    def _environment_remove(self, id_env):
        assert dnf.util.is_string_type(id_env) is True
        env_swdb = self.history.env.get(id_env)
        if not env_swdb:
            raise CompsError(_("Environment id '%s' is not installed.") % id_env)

        self.history.env.remove(env_swdb)

        tranbch = TransactionBunch()
        grid = set([i.getGroupId() for i in env_swdb.getGroups()])
        for id_grp in grid:
            if not self._removable_grp(id_grp):
                continue
            tranbch += self._group_remove(id_grp)
        return tranbch

    def _environment_install(self, id_env, typs_pkg, exclude=None, strict=True, groups_exclude=None):
        assert dnf.util.is_string_type(id_env)
        comps_env = self.comps._environment_by_id(id_env)
        if not comps_env:
            raise CompsError(_("Environment id '%s' does not exist.") % ucd(id_env))

        env_swdb = self.history.env.new(id_env, comps_env.name, comps_env.ui_name, typs_pkg)
        self.history.env.install(env_swdb)

        trans = TransactionBunch()
        for comps_group in comps_env.mandatory_groups:
            if groups_exclude and comps_group.id in groups_exclude:
                continue
            trans += self._group_install(comps_group.id, typs_pkg, exclude, strict)
            env_swdb.addGroup(comps_group.id, True, MANDATORY)

        for comps_group in comps_env.optional_groups:
            if groups_exclude and comps_group.id in groups_exclude:
                continue
            env_swdb.addGroup(comps_group.id, False, OPTIONAL)
            # TODO: if a group is already installed, mark it as installed?
        return trans


    def _environment_upgrade(self, id_env):
        assert dnf.util.is_string_type(id_env)
        env_swdb = self.history.env.get(id_env)
        env_cps = self.comps._environment_by_id(id_env)
        if not env_swdb:
            raise CompsError(_("Environment '%s' is not installed.") % id_env)
        if not env_cps:
            raise CompsError(_("Environment '%s' is not available.") % id_env)

        pkg_types = env_swdb.getPackageTypes()
        old_set = set([i.getGroupId() for i in env_swdb.getGroups()])

        # create a new record for current transaction
        env_swdb = self.history.env.new(env_cps.id, env_cps.name, env_cps.ui_name, pkg_types)

        trabch = TransactionBunch()
        for comps_group in env_cps.mandatory_groups:
            if comps_group.id in old_set:
                if self.history.group.get(comps_group.id):
                    # upgrade installed group
                    trabch += self._group_upgrade(comps_group.id)
            else:
                # install new group
                trabch += self._group_install(comps_group.id, pkg_types)
            env_swdb.addGroup(comps_group.id, True, MANDATORY)

        for comps_group in env_cps.optional_groups:
            if comps_group.id in old_set and self.history.group.get(comps_group.id):
                # upgrade installed group
                trabch += self._group_upgrade(comps_group.id)
            env_swdb.addGroup(comps_group.id, False, OPTIONAL)
            # TODO: if a group is already installed, mark it as installed?
        self.history.env.upgrade(env_swdb)
        return trabch
    
    def _group_remove(self, id_grp):
        assert dnf.util.is_string_type(id_grp)
        swdb_group = self.history.group.get(id_grp)
        if not swdb_group:
            raise CompsError(_("Module or Group '%s' is not installed.") % id_grp)
        self.history.group.remove(swdb_group)
        trans = TransactionBunch()
        trans.remove = {pkg for pkg in swdb_group.getPackages() if self._removable_pkg(pkg.getName())}
        return trans

    def _group_install(self, id_grp, typs_pkgs, exclude=None, strict=True, exclude_groups=None):
        assert dnf.util.is_string_type(id_grp)
        comps_group = self.comps._group_by_id(id_grp)
        if not comps_group:
            raise CompsError(_("Group id '%s' does not exist.") % ucd(id_grp))

        swdb_group = self.history.group.new(id_grp, comps_group.name, comps_group.ui_name, typs_pkgs)
        for i in comps_group.packages_iter():
            swdb_group.addPackage(i.name, False, Package._OPT_MAP[i.type])
        self.history.group.install(swdb_group)

        trans = TransactionBunch()
        # TODO: remove exclude
        if strict:
            trans.install.update(self._pkgs_of_type(comps_group, typs_pkgs, exclude=[]))
        else:
            trans.install_opt.update(self._pkgs_of_type(comps_group, typs_pkgs, exclude=[]))
        return trans

    def _group_upgrade(self, id_grp):
        assert dnf.util.is_string_type(id_grp)
        exclude = []
        swdb_group = self.history.group.get(id_grp)
        comps_group = self.comps._group_by_id(id_grp)

        if not swdb_group:
            argument = comps_group.ui_name if comps_group else id_grp
            raise CompsError(_("Module or Group '%s' is not installed.") % argument)
        if not comps_group:
            raise CompsError(_("Module or Group '%s' is not available.") % id_grp)
        pkg_types = swdb_group.getPackageTypes()
        set_new = self._pkgs_of_type(comps_group, pkg_types, exclude)
        set_old = set([i.getName() for i in swdb_group.getPackages()])

        # create a new record for current transaction
        swdb_group = self.history.group.new(id_grp, comps_group.name, comps_group.ui_name, pkg_types)
        for i in comps_group.packages_iter():
            swdb_group.addPackage(i.name, False, Package._OPT_MAP[i.type])
        self.history.group.upgrade(swdb_group)

        trans = TransactionBunch()
        trans.install = {pkg for pkg in set_new if pkg.name not in set_old}
        trans.remove = {name for name in set_old
                        if name not in [pkg.name for pkg in set_new]}
        trans.upgrade = {pkg for pkg in set_new if pkg.name in set_old}
        return trans

    def _exclude_packages_from_installed_groups(self, base):
        for grp in self.persistor.groups:
            p_grp = self.persistor.group(grp)
            if p_grp.installed:
                installed_pkg_names = \
                    set(p_grp.full_list) - set(p_grp.pkg_exclude)
                installed_pkgs = base.sack.query().installed().filterm(name=installed_pkg_names)
                for pkg in installed_pkgs:
                    base._goal.install(pkg)
