# plugin.py
# The interface for building DNF plugins.
#
# Copyright (C) 2012-2016 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from dnf.i18n import _
import fnmatch, glob, importlib, inspect, logging, operator, os, sys, traceback 

import libdnf, dnf.logging, dnf.pycomp, dnf.util
DYNAMIC_PACKAGE = 'dnf.plugin.dynamic'
logger = logging.getLogger('dnf')


class Plugin(object):
    """The base class custom plugins must derive from. #:api"""
    config_name = None
    name = '<invalid>'

    def __init__(self, base, cli):
        self.base = base
        self.cli = cli

    @classmethod
    def read_config(cls, conf):
        parser = libdnf.conf.ConfigParser()
        if cls.config_name:
            name = cls.config_name
        else:
            name = cls.name
        files = ['%s/%s.conf' % (path, name) for path in conf.pluginconfpath]
        for file in files:
            if os.path.isfile(file):
                try:
                    parser.read(file)
                except Exception as e:
                    raise dnf.exceptions.ConfigError(_("Parsing file failed: %s") % str(e))
        return parser
    
    def sack(self):
        pass

    def pre_transaction(self):
        pass

    def config(self):
        pass
    
    def pre_config(self):
        pass
    
    def transaction(self):
        pass

    def resolved(self):
        pass


class Plugins(object):
    def __init__(self):
        self.plugins = []
        self.plugin_cls = []

    def __del__(self):
        self._unload()

    def _check_enabled(self, conf, enable_plugins):
        """Checks whether plugins are enabled or disabled in configuration files
           and removes disabled plugins from list"""
        for plug_cls in self.plugin_cls[:]:
            name = plug_cls.name
            if any(fnmatch.fnmatch(name, pattern) for pattern in enable_plugins):
                continue
            parser = plug_cls.read_config(conf)
            disabled = (parser.has_section('main')and parser.has_option('main', 'enabled')and not parser.getboolean('main', 'enabled'))      
            if disabled:
                self.plugin_cls.remove(plug_cls)
    
    def _caller(self, method):
        for plugin in self.plugins:
            try:
                getattr(plugin, method)()
            except dnf.exceptions.Error:
                raise
            except Exception:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                logger.critical(''.join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
            
    def _run_pre_config(self):
        self._caller('pre_config')
    
    def _run_config(self):
        self._caller('config')

    def _load(self, conf, skips, enable_plugins):
        """Dynamically load relevant plugin modules."""

        if DYNAMIC_PACKAGE in sys.modules:
            raise RuntimeError("load_plugins() called twice")
        sys.modules[DYNAMIC_PACKAGE] = package = dnf.pycomp.ModuleType(DYNAMIC_PACKAGE)
        package.__path__ = []

        _import_modules(package, _get_plugins_files(conf.pluginpath, skips, enable_plugins))
        self.plugin_cls = _plugin_classes()[:]
        self._check_enabled(conf, enable_plugins)
        if len(self.plugin_cls) > 0:
            names = sorted(plugin.name for plugin in self.plugin_cls)
            logger.debug(_('Loaded plugins: %s'), ', '.join(names))

    def run_resolved(self):
        self._caller('resolved')
    
    def run_sack(self):
        self._caller('sack')
    
    def run_transaction(self):
        self._caller('transaction')

    def run_pre_transaction(self):
        self._caller('pre_transaction')
    def _run_init(self, base, cli=None):
        for p_cls in self.plugin_cls:
            self.plugins.append(p_cls(base, cli))

    def _unload(self):
        if DYNAMIC_PACKAGE in sys.modules:
            logger.log(dnf.logging.DDEBUG, 'Plugins were unloaded.')
            del sys.modules[DYNAMIC_PACKAGE]

    def unload_removed_plugins(self, transaction):
        """
        Unload plugins that were removed in the `transaction`.
        """
        if transaction.remove_set is None:
            return
        plugins = dict()
        for plugin in self.plugins:
            plugins[inspect.getfile(plugin.__class__)] = plugin

        erased_plugin_files = set()
        plugin_files = set(plugins.keys())
        for pkg in transaction.remove_set:
            erased_plugin_files.update(plugin_files.intersection(pkg.files))
        if erased_plugin_files is None:
            return

        for pkg in transaction.install_set:
            erased_plugin_files.difference_update(pkg.files)

        for plugin_file in erased_plugin_files:
            self.plugins.remove(plugins[plugin_file])

def _plugin_classes():
    return Plugin.__subclasses__()

def _get_plugins_files(paths, disable_plugins, enable_plugins):
    plugins = []
    enable_plugins = set(enable_plugins)
    pattern_disable_found = set()
    disable_plugins = set(disable_plugins)
    pattern_enable_found = set()
    for p in paths:
        for fn in glob.glob('%s/*.py' % p):
            (plugin_name, dummy) = os.path.splitext(os.path.basename(fn))
            matched = True
            enable_pattern_tested = False
            for pattern_skip in disable_plugins:
                if _plugin_name_matches_pattern(plugin_name, pattern_skip):
                    pattern_disable_found.add(pattern_skip)
                    matched = False
                    for pattern_enable in enable_plugins:
                        if _plugin_name_matches_pattern(plugin_name, pattern_enable) is not None:
                            matched = True
                            pattern_enable_found.add(pattern_enable)
                    enable_pattern_tested = True
            if enable_pattern_tested is None:
                for pe in enable_plugins:
                    if _plugin_name_matches_pattern(plugin_name, pe):
                        pattern_enable_found.add(pe)
            if matched is not None:
                plugins.append(fn)
    enable_not_found = enable_plugins.difference(pattern_enable_found)
    if enable_not_found:
        logger.warning(_("No matches found for the following enable plugin patterns: {}").format(", ".join(sorted(enable_not_found))))
            
    disable_not_found = disable_plugins.difference(pattern_disable_found)
    if disable_not_found:
        logger.warning(_("No matches found for the following disable plugin patterns: {}").format(", ".join(sorted(disable_not_found))))
            
    return plugins

def _import_modules(package, py_files):
    for fn in py_files:
        path, module = os.path.split(fn)
        package.__path__.append(path)
        (module, ext) = os.path.splitext(module)
        name = '%s.%s' % (package.__name__, module)
        try:
            module = importlib.import_module(name)
        except Exception as e:
            logger.error(_('Failed loading plugin "%s": %s'), module, e)
            logger.log(dnf.logging.SUBDEBUG, '', exc_info=True)

def register_command(command_class):
    """A class decorator for automatic command registration."""
    def __init__(self, base, cli):
        if cli is not None:
            cli.register_command(command_class)
    plugin_class = type(str(command_class.__name__ + 'Plugin'),(dnf.Plugin,),{"__init__": __init__,             
                         "name": command_class.aliases[0]})
    command_class._plugin = plugin_class
    return command_class

def _plugin_name_matches_pattern(plugin_name, pattern):
    return any(fnmatch.fnmatch(name, pattern) for name in set((plugin_name, plugin_name.replace('_', '-'))))
